#include <string.h>
#include "pdblib.h"

static int _removeHOH(PDBhandle PDBh, PDBlib_iter_data_t *data) {
	if(strcmp(data->r->name, "HOH") == 0 || strcmp(data->r->name, "WAT") == 0){
		PDBlib_del_residue(PDBh, data->r->id);
		return 2;}
	return 0;
}

/*! @ingroup PDBlib
 * @brief Remove water atoms from structure.
 * @param PDBh - initialized PDBlib handler
 *
 * This function removes from PDB structure all water atoms. Water is recognized
 * as residue with "HOH" or "WAT" name.
 */
void PDBlib_remove_water(PDBhandle PDBh)
{
	PDBlib_iterator(PDBh, PDBLIB_ITER_RESIDUE, _removeHOH, NULL);
}

static int _removeHetero(PDBhandle PDBh, PDBlib_iter_data_t *data) {
	PDBlib_atom_t *cur_a;
	int xa;

	// if every atoms in residue is hetero atom -> del residue
	if(data->r->atoms->len == data->r->hetero_counter){
		PDBlib_del_residue(PDBh, data->r->id);
		return 2;}

	// remove single heteroaotms form residue
	for(xa = 0; xa < data->r->atoms->len; xa++){
		cur_a = &g_array_index(data->r->atoms, PDBlib_atom_t, xa);
		if(cur_a->hetero == 1) PDBlib_del_atom(PDBh, xa--);}
	return 0;
}

/*! @ingroup PDBlib
 * @brief Remove heteroatoms from structure.
 * @param PDBh - initialized PDBlib handler
 *
 * This function removes from PDB structure all atoms which are marked as
 * heteroatoms. If in particular residue are only such atoms then whole
 * residue is removed.
 */
void PDBlib_remove_heteroatoms(PDBhandle PDBh)
{
	PDBlib_iterator(PDBh, PDBLIB_ITER_RESIDUE, _removeHetero, NULL);
}

/*! @ingroup PDBlib
 * @brief Set atom ID to hash value.
 * @param src_a - pointer to atom structure
 * @return Unique hashed atom ID
 *
 * Atom ID is set to hash value based on:
 * @n + chain:   name
 * @n + residue: name, sequence number
 * @n + atom:    name, hetero bit, alternative location
 *
 * After this operation atom ID should be unique for whole PDB structure.
 */
int PDBlib_set_atom_hashed_id(PDBlib_atom_t *src_a)
{
	PDBlib_chain_t *cur_c;
	PDBlib_residue_t *cur_r;
	int hashed_id, x;

	cur_r = src_a->top_residue;
	cur_c = cur_r->top_chain;

	for(x = hashed_id = 0; x < strlen(src_a->name); x++)
		hashed_id += src_a->name[x]*(x + src_a->name[x]/6)*
				(x + (6 + src_a->hetero)*src_a->name[x]);
	hashed_id *= cur_r->sequence_nb + src_a->alt_location;
	hashed_id += cur_c->name[0];
	for(x = 0; x < strlen(cur_r->name); x++)
		hashed_id += cur_r->name[x]*(x + 9)*(x + 9 + src_a->hetero);

	return src_a->id = hashed_id;
}

/*! @ingroup PDBlib
 * @brief Set residue ID to hash value.
 * @param src_r - pointer to residue structure
 * @return Unique hashed residue ID
 *
 * Residue ID is set to hash value based on:
 * @n + chain:   name
 * @n + residue: name, sequence number
 */
int PDBlib_set_residue_hashed_id(PDBlib_residue_t *src_r)
{
	PDBlib_chain_t *cur_c;
	int hashed_id, x;

	cur_c = src_r->top_chain;

	for(x = hashed_id = 0; x < strlen(src_r->name); x++)
		hashed_id += src_r->name[x]*(x + 9)*(x + 9);
	hashed_id += cur_c->name[0];
	hashed_id *= src_r->sequence_nb;

	return src_r->id = hashed_id;
}

/*! @ingroup PDBlib
 * @brief Copy residue data from src to dst.
 * @param dst_r - pointer to destination residue structure
 * @param src_r - pointer to source residue structure
 *
 * You should use this function if you want to copy all data from one residue
 * structure to another. This function preserves all crucial information
 * for PDBhandle consistence associated with destination residue structure.
 */
void PDBlib_copy_residue_data(PDBlib_residue_t *dst_r, const PDBlib_residue_t *src_r)
{
	dst_r->sequence_nb = src_r->sequence_nb;
	dst_r->hetero_counter = src_r->hetero_counter;
	dst_r->insertion = src_r->insertion;
	memcpy(dst_r->name, src_r->name, sizeof(dst_r->name));
	memcpy(dst_r->segment, src_r->segment, sizeof(dst_r->segment));
}

/*! @ingroup PDBlib
 * @brief Copy atom data from src to dst.
 * @param dst_a - pointer to destination atom structure
 * @param src_a - pointer to source atom structure
 *
 * You should use this function if you want to copy all data from one atom
 * structure to another. This function preserves all crucial information
 * for PDBhandle consistence associated with destination atom structure.
 */
void PDBlib_copy_atom_data(PDBlib_atom_t *dst_a, const PDBlib_atom_t *src_a)
{
	dst_a->serial_nb = src_a->serial_nb; dst_a->hetero = src_a->hetero;
	dst_a->coordXYZ[0] = src_a->coordXYZ[0]; dst_a->coordXYZ[1] = src_a->coordXYZ[1];
	dst_a->coordXYZ[2] = src_a->coordXYZ[2]; dst_a->occupancy = src_a->occupancy;
	dst_a->bfactor = src_a->bfactor; dst_a->alt_location = src_a->alt_location;
	memcpy(dst_a->name, src_a->name, sizeof(dst_a->name));
	memcpy(dst_a->element, src_a->element, sizeof(dst_a->element));
	memcpy(dst_a->charge, src_a->charge, sizeof(dst_a->charge));
	memcpy(dst_a->aniso_temp_factor, src_a->aniso_temp_factor, sizeof(dst_a->aniso_temp_factor));
}

/*! @ingroup PDBlib
 * @brief Make new copy of PDB structure
 * @param PDBh - initialized PDBlib handler
 * @return PDBhandle with copied structure
 *
 * This function makes exact copy of source PDB structure. Note that returned
 * handler has to be freed with a call to PDBlib_free().
 */
PDBhandle PDBlib_duplicate(const PDBhandle PDBh)
{
	PDBhandle PDBh_new;
	PDBlib_iter_data_t objs_old, objs_new;
	int xs, xm, xc, xr, xa;
	int xss;

	if((PDBh_new = PDBlib_initialize()) == NULL) return NULL;

	// iterate structures
	for(xs = 0; xs < PDBh->structures->len; xs++){
		objs_old.s = &g_array_index(PDBh->structures, PDBlib_struct_t, xs);
		objs_new.s = PDBlib_add_structure(PDBh_new);

		// copy structure data
		objs_new.s->id = objs_old.s->id;
		memcpy(objs_new.s->classification, objs_old.s->classification, sizeof(objs_new.s->classification));
		memcpy(objs_new.s->PDBID, objs_old.s->PDBID, sizeof(objs_new.s->PDBID));
		memcpy(&objs_new.s->depos_data, &objs_old.s->depos_data, sizeof(objs_new.s->depos_data));

		// iterate ssbonds
		for(xss = 0; xss < objs_old.s->ssbonds->len; xss++) {
			objs_old.sb = &g_array_index(objs_old.s->ssbonds, PDBlib_ssbond_t, xss);
			objs_new.sb = PDBlib_add_ssbond(PDBh_new);

			// copy ssbond data
			objs_new.sb->id = objs_old.sb->id;
			objs_new.sb->serial_nb = objs_old.sb->serial_nb;
			objs_new.sb->resseq_nb1 = objs_old.sb->resseq_nb1;
			objs_new.sb->resseq_nb2 = objs_old.sb->resseq_nb2;
			objs_new.sb->sym_oper1 = objs_old.sb->sym_oper1;
			objs_new.sb->sym_oper2 = objs_old.sb->sym_oper2;
			objs_new.sb->res_insertion1 = objs_old.sb->res_insertion1;
			objs_new.sb->res_insertion2 = objs_old.sb->res_insertion2;
			objs_new.sb->sslen = objs_old.sb->sslen;
			memcpy(objs_new.sb->chainame1, objs_old.sb->chainame1, sizeof(objs_new.sb->chainame1));
			memcpy(objs_new.sb->chainame2, objs_old.sb->chainame2, sizeof(objs_new.sb->chainame2));
		}

		// iterate models
		for(xm = 0; xm < objs_old.s->models->len; xm++) {
			objs_old.m = &g_array_index(objs_old.s->models, PDBlib_model_t, xm);
			objs_new.m = PDBlib_add_model(PDBh_new);

			// copy model data
			objs_new.m->id = objs_old.m->id;
			objs_new.m->serial_nb = objs_old.m->serial_nb;

			// iterate chains
			for(xc = 0; xc < objs_old.m->chains->len; xc++) {
				objs_old.c = &g_array_index(objs_old.m->chains, PDBlib_chain_t, xc);
				objs_new.c = PDBlib_add_chain(PDBh_new);

				// copy chain data
				objs_new.c->id = objs_old.c->id;
				memcpy(objs_new.c->name, objs_old.c->name, sizeof(objs_new.c->name));

				// iterate residues
				for(xr = 0; xr < objs_old.c->residues->len; xr++) {
					objs_old.r = &g_array_index(objs_old.c->residues, PDBlib_residue_t, xr);
					objs_new.r = PDBlib_add_residue(PDBh_new);

					objs_new.r->id = objs_old.r->id;
					PDBlib_copy_residue_data(objs_new.r, objs_old.r);

					// iterate atoms
					for(xa = 0; xa < objs_old.r->atoms->len; xa++) {
						objs_old.a = &g_array_index(objs_old.r->atoms, PDBlib_atom_t, xa);
						objs_new.a = PDBlib_add_atom(PDBh_new);

						objs_new.a->id = objs_old.a->id;
						PDBlib_copy_atom_data(objs_new.a, objs_old.a);

	}}}}}

	return PDBh_new;
}
