#include <string.h>
#include "pdblib.h"

/*! @ingroup PDBlib
 * @brief Iteration over specified objects.
 * @param PDBh - initialized PDBlib handler
 * @param type - iteration type
 * @param callback - callback function
 * @param userdata - pointer to user data
 *
 * Defined iteration types:
 * @n + PDBLIB_ITER_ATOM - iteration over all atoms
 * @n + PDBLIB_ITER_RESIDUE - iteration over all residues
 * @n + PDBLIB_ITER_CHAIN - iteration over all chains
 * @n + PDBLIB_ITER_MODEL - iteration over all models
 * @n + PDBLIB_ITER_STRUCTURE - iteration over all structures
 * @n + PDBLIB_ITER_SSBOND - iteration over all S-S bonds
 *
 * Invoke callback function for every object in PDB structure. Iteration is
 * breaking if callback function returns 1. Redo the same iteration if
 * callback returns 2 (e.g. after object deletion).
 */
void PDBlib_iterator(PDBhandle PDBh, int type, PDBlib_iter_callback_t callback, void *userdata)
{
	PDBlib_iter_data_t objs;
	int xs, xm, xc, xr, xa;
	int xss;

	bzero(&objs, sizeof(objs));
	objs.userdata = userdata;
	bzero(&PDBh->c, sizeof(PDBh->c));

	// iterate structures
	for(xs = 0; xs < PDBh->structures->len; xs++){
		objs.s = &g_array_index(PDBh->structures, PDBlib_struct_t, xs);
		PDBh->c.s = objs.s;

		if(type == PDBLIB_ITER_STRUCTURE) {
			switch(callback(PDBh, &objs)){
			case 1: return;
			case 2: xs--;}

			continue;
		}

		// iterate ssbonds
		if(type == PDBLIB_ITER_SSBOND) {
			for(xss = 0; xss < objs.s->ssbonds->len; xss++) {
				objs.sb = &g_array_index(objs.s->ssbonds, PDBlib_ssbond_t, xss);
				PDBh->c.sb = objs.sb;

				switch(callback(PDBh, &objs)){
				case 1: return;
				case 2: xss--;}
			}
			continue;
		}


		// iterate models
		for(xm = 0; xm < objs.s->models->len; xm++) {
			objs.m = &g_array_index(objs.s->models, PDBlib_model_t, xm);
			PDBh->c.m = objs.m;

			if(type == PDBLIB_ITER_MODEL) {
				switch(callback(PDBh, &objs)){
				case 1: return;
				case 2: xm--;}

				continue;
			}

			// iterate chains
			for(xc = 0; xc < objs.m->chains->len; xc++) {
				objs.c = &g_array_index(objs.m->chains, PDBlib_chain_t, xc);
				PDBh->c.c = objs.c;

				if(type == PDBLIB_ITER_CHAIN) {
					switch(callback(PDBh, &objs)){
					case 1: return;
					case 2: xc--;}

					continue;
				}

				// iterate residues
				for(xr = 0; xr < objs.c->residues->len; xr++) {
					objs.r = &g_array_index(objs.c->residues, PDBlib_residue_t, xr);
					PDBh->c.r = objs.r;

					if(type == PDBLIB_ITER_RESIDUE) {
						switch(callback(PDBh, &objs)){
						case 1: return;
						case 2: xr--;}

						continue;
					}

					// iterate atoms
					for(xa = 0; xa < objs.r->atoms->len; xa++) {
						objs.a = &g_array_index(objs.r->atoms, PDBlib_atom_t, xa);
						PDBh->c.a = objs.a;

						if(type == PDBLIB_ITER_ATOM)
							switch(callback(PDBh, &objs)){
							case 1: return;
							case 2: xa--;}

	}}}}}
}

