#include <stdlib.h>
#include <string.h>
#include "pdblib.h"

/*! @ingroup PDBlib
 * @brief Add new structure object.
 * @param PDBh - initialized PDBlib handler
 * @return Functions returns pointer to PDBlib_struct_t structure.
 *
 * Newly created structure object is set as referencing point for
 * various operations on PDB hierarchy.
 */
PDBlib_struct_t *PDBlib_add_structure(PDBhandle PDBh)
{
	PDBlib_struct_t new_s, *cur_s;
	PDBlib_model_t *cur_m;
	void *garray_data_ptr;
	int xs, xm;

	bzero(&new_s, sizeof(new_s));
	new_s.id = random() + 1;
	new_s.models = g_array_new(0, 0, sizeof(PDBlib_model_t));
	new_s.ssbonds = g_array_new(0, 0, sizeof(PDBlib_ssbond_t));

	garray_data_ptr = PDBh->structures->data;
	g_array_append_vals(PDBh->structures, &new_s, 1);
	bzero(&PDBh->c, sizeof(PDBh->c));

	if(garray_data_ptr != PDBh->structures->data) //GArray reallocation
		// update top-references in models
		for(xs = 0; xs < PDBh->structures->len; xs++) {
			cur_s = &g_array_index(PDBh->structures, PDBlib_struct_t, xs);
			for(xm = 0; xm < cur_s->models->len; xm++){
				cur_m = &g_array_index(cur_s->models, PDBlib_model_t, xm);
				cur_m->top_structure = cur_s;}
		}

	return PDBh->c.s = &g_array_index(PDBh->structures, PDBlib_struct_t,
			PDBh->structures->len - 1);
}

/*! @ingroup PDBlib
 * @brief Get structure object.
 * @param PDBh - initialized PDBlib handler
 * @param id - unique structure ID
 * @return On success pointer to PDBlib_struct_t structure is returned.
 *
 * Function searches for structure object with given ID. If such a structure
 * exists then pointer to this structure is returned, otherwise function
 * returns NULL. If succeed, founded structure object is set as referencing
 * point.
 */
PDBlib_struct_t *PDBlib_get_structure(PDBhandle PDBh, int id)
{
	PDBlib_struct_t *cur_s;
	int xs;

	for(xs = 0; xs < PDBh->structures->len; xs++) {
		cur_s = &g_array_index(PDBh->structures, PDBlib_struct_t, xs);
		if(cur_s->id == id){
			bzero(&PDBh->c, sizeof(PDBh->c));
			return PDBh->c.s = cur_s;}
	}
	return NULL;
}

/*! @ingroup PDBlib
 * @brief Delete structure object.
 * @param PDBh - initialized PDBlib handler
 * @param id - unique structure ID
 * @return On success 0 is returned, otherwise -1.
 *
 * Function searches for structure object with given ID. If such a structure
 * exists then this structure and all underlaying objects are deleted.
 * Also memory associated with these objects is freed and all referencing
 * points are cleared.
 */
int PDBlib_del_structure(PDBhandle PDBh, int id)
{
	PDBlib_struct_t *cur_s = NULL;
	PDBlib_model_t *cur_m;
	int xs, xm;

	for(xs = 0; xs < PDBh->structures->len; xs++){
		cur_s = &g_array_index(PDBh->structures, PDBlib_struct_t, xs);
		if(cur_s->id == id) break;}

	// structure with requested ID not found
	if(xs == PDBh->structures->len) return -1;

	// delete underlaying models and free models handler
	for(xm = 0; xm < cur_s->models->len; xm++){
		cur_m = &g_array_index(cur_s->models, PDBlib_model_t, xm);
		PDBlib_del_model(PDBh, cur_m->id);}
	g_array_free(cur_s->models, TRUE);

	// free underlaying ssbonds handler
	g_array_free(cur_s->ssbonds, TRUE);

	g_array_remove_index(PDBh->structures, xs);

	// update top-references in models of remaining structures
	for(; xs < PDBh->structures->len; xs++) {
		cur_s = &g_array_index(PDBh->structures, PDBlib_struct_t, xs);
		for(xm = 0; xm < cur_s->models->len; xm++){
			cur_m = &g_array_index(cur_s->models, PDBlib_model_t, xm);
			cur_m->top_structure = cur_s;}
	}

	bzero(&PDBh->c, sizeof(PDBh->c));
	return 0;
}

/*! @ingroup PDBlib
 * @brief Add new ssbond object.
 * @param PDBh - initialized PDBlib handler
 * @return Functions returns pointer to PDBlib_ssbond_t structure.
 *
 * Newly created ssbond object is set as referencing point.
 */
PDBlib_ssbond_t *PDBlib_add_ssbond(PDBhandle PDBh)
{
	PDBlib_ssbond_t new_ss;

	// if current structure object is NULL add new one
	if(PDBh->c.s == NULL) PDBh->c.s = PDBlib_add_structure(PDBh);

	bzero(&new_ss, sizeof(new_ss));
	new_ss.id = random() + 1;
	new_ss.top_structure = PDBh->c.s;

	g_array_append_vals(PDBh->c.s->ssbonds, &new_ss, 1);
	return PDBh->c.sb = &g_array_index(PDBh->c.s->ssbonds, PDBlib_ssbond_t,
			PDBh->c.s->ssbonds->len - 1);
}

/*! @ingroup PDBlib
 * @brief Get ssbond object.
 * @param PDBh - initialized PDBlib handler
 * @param id - unique ssbond ID
 * @return On success pointer to PDBlib_ssbond_t structure is returned.
 *
 * For more information see PDBlib_get_structure().
 */
PDBlib_ssbond_t *PDBlib_get_ssbond(PDBhandle PDBh, int id)
{
	PDBlib_ssbond_t *cur_ss;
	int xss;

	if(PDBh->c.s == NULL) return NULL;
	for(xss = 0; xss < PDBh->c.s->ssbonds->len; xss++){
		cur_ss = &g_array_index(PDBh->c.s->ssbonds, PDBlib_ssbond_t, xss);
		if(cur_ss->id == id) return PDBh->c.sb = cur_ss;}
	return NULL;
}

/*! @ingroup PDBlib
 * @brief Delete ssbond object.
 * @param PDBh - initialized PDBlib handler
 * @param id - unique ssbond ID
 * @return On success 0 is returned, otherwise -1.
 *
 * For more information see PDBlib_del_structure().
 */
int PDBlib_del_ssbond(PDBhandle PDBh, int id)
{
	PDBlib_ssbond_t *cur_ss;
	int xss;

	if(PDBh->c.s == NULL) return -1;
	for(xss = 0; xss < PDBh->c.s->ssbonds->len; xss++){
		cur_ss = &g_array_index(PDBh->c.s->ssbonds, PDBlib_ssbond_t, xss);
		if(cur_ss->id == id) break;}

	// ssbond with requested ID not found
	if(xss == PDBh->c.s->ssbonds->len) return -1;

	g_array_remove_index(PDBh->c.s->ssbonds, xss);

	PDBh->c.sb = NULL;
	return 0;
}

/*! @ingroup PDBlib
 * @brief Add new model object.
 * @param PDBh - initialized PDBlib handler
 * @return Functions returns pointer to PDBlib_model_t structure.
 *
 * Function creates new model object and preallocates space for 3 chain
 * objects within this model. If needed top-level PDBlib_struct_t object is
 * created as well. Newly created model object is set as referencing point.
 */
PDBlib_model_t *PDBlib_add_model(PDBhandle PDBh)
{
	PDBlib_model_t new_m, *cur_m;
	PDBlib_chain_t *cur_c;
	void *garray_data_ptr;
	int xm, xc;

	// if current structure object is NULL add new one
	if(PDBh->c.s == NULL) PDBh->c.s = PDBlib_add_structure(PDBh);

	bzero(&new_m, sizeof(new_m));
	new_m.id = random() + 1;
	new_m.top_structure = PDBh->c.s;
	new_m.chains = g_array_sized_new(0, 0, sizeof(PDBlib_chain_t), 3);

	garray_data_ptr = PDBh->c.s->models->data;
	g_array_append_vals(PDBh->c.s->models, &new_m, 1);
	PDBh->c.c = NULL; PDBh->c.r = NULL; PDBh->c.a = NULL;

	if(garray_data_ptr != PDBh->c.s->models->data) //GArray reallocation
		// update top-references in chains
		for(xm = 0; xm < PDBh->c.s->models->len; xm++) {
			cur_m = &g_array_index(PDBh->c.s->models, PDBlib_model_t, xm);
			for(xc = 0; xc < cur_m->chains->len; xc++){
				cur_c = &g_array_index(cur_m->chains, PDBlib_chain_t, xc);
				cur_c->top_model = cur_m;}
		}

	return PDBh->c.m = &g_array_index(PDBh->c.s->models, PDBlib_model_t,
			PDBh->c.s->models->len - 1);
}

/*! @ingroup PDBlib
 * @brief Get model object.
 * @param PDBh - initialized PDBlib handler
 * @param id - unique model ID
 * @return On success pointer to PDBlib_model_t structure is returned.
 *
 * For more information see PDBlib_get_structure().
 */
PDBlib_model_t *PDBlib_get_model(PDBhandle PDBh, int id)
{
	PDBlib_model_t *cur_m;
	int xm;

	if(PDBh->c.s == NULL) return NULL;
	for(xm = 0; xm < PDBh->c.s->models->len; xm++) {
		cur_m = &g_array_index(PDBh->c.s->models, PDBlib_model_t, xm);
		if(cur_m->id == id){
			PDBh->c.c = NULL; PDBh->c.r = NULL; PDBh->c.a = NULL;
			return PDBh->c.m = cur_m;}
	}
	return NULL;
}

/*! @ingroup PDBlib
 * @brief Delete model object.
 * @param PDBh - initialized PDBlib handler
 * @param id - unique model ID
 * @return On success 0 is returned, otherwise -1.
 *
 * For more information see PDBlib_del_structure().
 */
int PDBlib_del_model(PDBhandle PDBh, int id)
{
	PDBlib_model_t *cur_m = NULL;
	PDBlib_chain_t *cur_c;
	PDBlib_residue_t *cur_r;
	int xm, xc, xr;

	if(PDBh->c.s == NULL) return -1;
	for(xm = 0; xm < PDBh->c.s->models->len; xm++){
		cur_m = &g_array_index(PDBh->c.s->models, PDBlib_model_t, xm);
		if(cur_m->id == id) break;}

	// model with requested ID not found
	if(xm == PDBh->c.s->models->len) return -1;

	// free underlaying chains, residues and atoms handlers
	for(xc = 0; xc < cur_m->chains->len; xc++){
		cur_c = &g_array_index(cur_m->chains, PDBlib_chain_t, xc);
		for(xr = 0; xr < cur_c->residues->len; xr++){
			cur_r = &g_array_index(cur_c->residues, PDBlib_residue_t, xr);
			g_array_free(cur_r->atoms, TRUE);}
		g_array_free(cur_c->residues, TRUE);}
	g_array_free(cur_m->chains, TRUE);

	g_array_remove_index(PDBh->c.s->models, xm);

	// update top-references in chains of remaining models
	for(; xm < PDBh->c.s->models->len; xm++) {
		cur_m = &g_array_index(PDBh->c.s->models, PDBlib_model_t, xm);
		for(xc = 0; xc < cur_m->chains->len; xc++){
			cur_c = &g_array_index(cur_m->chains, PDBlib_chain_t, xc);
			cur_c->top_model = cur_m;}
	}

	PDBh->c.m = NULL; PDBh->c.c = NULL; PDBh->c.r = NULL; PDBh->c.a = NULL;
	return 0;
}

/*! @ingroup PDBlib
 * @brief Add new chain object.
 * @param PDBh - initialized PDBlib handler
 * @return Functions returns pointer to PDBlib_chain_t structure.
 *
 * Function creates new chain object and preallocates space for 150
 * residues objects within this chain. If needed top-level objects
 * are created as well. Newly created chain object is set as referencing
 * point.
 */
PDBlib_chain_t *PDBlib_add_chain(PDBhandle PDBh)
{
	PDBlib_chain_t new_c, *cur_c;
	PDBlib_residue_t *cur_r;
	void *garray_data_ptr;
	int xc, xr;

	// if current model object is NULL add new one
	if(PDBh->c.m == NULL) PDBh->c.m = PDBlib_add_model(PDBh);

	bzero(&new_c, sizeof(new_c));
	new_c.id = random() + 1;
	new_c.top_model = PDBh->c.m;
	new_c.residues = g_array_sized_new(0, 0, sizeof(PDBlib_residue_t), 150);

	garray_data_ptr = PDBh->c.m->chains->data;
	g_array_append_vals(PDBh->c.m->chains, &new_c, 1);
	PDBh->c.r = NULL; PDBh->c.a = NULL;

	if(garray_data_ptr != PDBh->c.m->chains->data) //GArray reallocation
		// update top-references in residues
		for(xc = 0; xc < PDBh->c.m->chains->len; xc++) {
			cur_c = &g_array_index(PDBh->c.m->chains, PDBlib_chain_t, xc);
			for(xr = 0; xr < cur_c->residues->len; xr++){
				cur_r = &g_array_index(cur_c->residues, PDBlib_residue_t, xr);
				cur_r->top_chain = cur_c;}
		}

	return PDBh->c.c = &g_array_index(PDBh->c.m->chains, PDBlib_chain_t,
			PDBh->c.m->chains->len - 1);
}

/*! @ingroup PDBlib
 * @brief Get chain object.
 * @param PDBh - initialized PDBlib handler
 * @param id - unique chain ID
 * @return On success pointer to PDBlib_chain_t structure is returned.
 *
 * For more information see PDBlib_get_structure().
 */
PDBlib_chain_t *PDBlib_get_chain(PDBhandle PDBh, int id)
{
	PDBlib_chain_t *cur_c;
	int xc;

	if(PDBh->c.m == NULL) return NULL;
	for(xc = 0; xc < PDBh->c.m->chains->len; xc++) {
		cur_c = &g_array_index(PDBh->c.m->chains, PDBlib_chain_t, xc);
		if(cur_c->id == id){
			PDBh->c.r = NULL; PDBh->c.a = NULL;
			return PDBh->c.c = cur_c;}
	}
	return NULL;
}

/*! @ingroup PDBlib
 * @brief Delete chain object.
 * @param PDBh - initialized PDBlib handler
 * @param id - unique chain ID
 * @return On success 0 is returned, otherwise -1.
 *
 * For more information see PDBlib_del_structure().
 */
int PDBlib_del_chain(PDBhandle PDBh, int id)
{
	PDBlib_chain_t *cur_c = NULL;
	PDBlib_residue_t *cur_r;
	int xc, xr;

	if(PDBh->c.m == NULL) return -1;
	for(xc = 0; xc < PDBh->c.m->chains->len; xc++){
		cur_c = &g_array_index(PDBh->c.m->chains, PDBlib_chain_t, xc);
		if(cur_c->id == id) break;}

	// chain with requested ID not found
	if(xc == PDBh->c.m->chains->len) return -1;

	// free underlaying residues and atoms handlers
	for(xr = 0; xr < cur_c->residues->len; xr++){
		cur_r = &g_array_index(cur_c->residues, PDBlib_residue_t, xr);
		g_array_free(cur_r->atoms, TRUE);}
	g_array_free(cur_c->residues, TRUE);

	g_array_remove_index(PDBh->c.m->chains, xc);

	// update top-references in residues of remaining chains
	for(; xc < PDBh->c.m->chains->len; xc++) {
		cur_c = &g_array_index(PDBh->c.m->chains, PDBlib_chain_t, xc);
		for(xr = 0; xr < cur_c->residues->len; xr++){
			cur_r = &g_array_index(cur_c->residues, PDBlib_residue_t, xr);
			cur_r->top_chain = cur_c;}
	}

	PDBh->c.c = NULL; PDBh->c.r = NULL; PDBh->c.a = NULL;
	return 0;
}

/*! @ingroup PDBlib
 * @brief Add new residue object.
 * @param PDBh - initialized PDBlib handler
 * @return Functions returns pointer to PDBlib_residue_t structure.
 *
 * Function creates new residue object and preallocates space for 10
 * atoms objects within this residue. If needed top-level objects
 * are created as well. Newly created residue object is set as
 * referencing point.
 */
PDBlib_residue_t *PDBlib_add_residue(PDBhandle PDBh)
{
	PDBlib_residue_t new_r, *cur_r;
	PDBlib_atom_t *cur_a;
	void *garray_data_ptr;
	int xr, xa;

	// if current chain object is NULL add new one
	if(PDBh->c.c == NULL) PDBh->c.c = PDBlib_add_chain(PDBh);

	bzero(&new_r, sizeof(new_r));
	new_r.id = random() + 1;
	new_r.top_chain = PDBh->c.c;
	new_r.atoms = g_array_sized_new(0, 0, sizeof(PDBlib_atom_t), 10);

	garray_data_ptr = PDBh->c.c->residues->data;
	g_array_append_vals(PDBh->c.c->residues, &new_r, 1);
	PDBh->c.a = NULL;

	if(garray_data_ptr != PDBh->c.c->residues->data) //GArray reallocation
		// update top-references in atoms
		for(xr = 0; xr < PDBh->c.c->residues->len; xr++) {
			cur_r = &g_array_index(PDBh->c.c->residues, PDBlib_residue_t, xr);
			for(xa = 0; xa < cur_r->atoms->len; xa++){
				cur_a = &g_array_index(cur_r->atoms, PDBlib_atom_t, xa);
				cur_a->top_residue = cur_r;}
		}

	return PDBh->c.r = &g_array_index(PDBh->c.c->residues, PDBlib_residue_t,
			PDBh->c.c->residues->len - 1);
}

/*! @ingroup PDBlib
 * @brief Get residue object.
 * @param PDBh - initialized PDBlib handler
 * @param id - unique residue ID
 * @return On success pointer to PDBlib_residue_t structure is returned.
 *
 * For more information see PDBlib_get_structure().
 */
PDBlib_residue_t *PDBlib_get_residue(PDBhandle PDBh, int id)
{
	PDBlib_residue_t *cur_r;
	int xr;

	if(PDBh->c.c == NULL) return NULL;
	for(xr = 0; xr < PDBh->c.c->residues->len; xr++) {
		cur_r = &g_array_index(PDBh->c.c->residues, PDBlib_residue_t, xr);
		if(cur_r->id == id){
			PDBh->c.a = NULL;
 			return PDBh->c.r = cur_r;}
	}
	return NULL;
}

/*! @ingroup PDBlib
 * @brief Delete residue object.
 * @param PDBh - initialized PDBlib handler
 * @param id - unique residue ID
 * @return On success 0 is returned, otherwise -1.
 *
 * For more information see PDBlib_del_structure().
 */
int PDBlib_del_residue(PDBhandle PDBh, int id)
{
	PDBlib_residue_t *cur_r = NULL;
	PDBlib_atom_t *cur_a;
	int xr, xa;

	if(PDBh->c.c == NULL) return -1;
	for(xr = 0; xr < PDBh->c.c->residues->len; xr++){
		cur_r = &g_array_index(PDBh->c.c->residues, PDBlib_residue_t, xr);
		if(cur_r->id == id) break;}

	// residue with requested ID not found
	if(xr == PDBh->c.c->residues->len) return -1;

	// free underlaying atoms handler
	g_array_free(cur_r->atoms, TRUE);

	g_array_remove_index(PDBh->c.c->residues, xr);

	// update top-references in atoms of remaining residues
	for(; xr < PDBh->c.c->residues->len; xr++) {
		cur_r = &g_array_index(PDBh->c.c->residues, PDBlib_residue_t, xr);
		for(xa = 0; xa < cur_r->atoms->len; xa++){
			cur_a = &g_array_index(cur_r->atoms, PDBlib_atom_t, xa);
			cur_a->top_residue = cur_r;}
	}

	PDBh->c.r = NULL; PDBh->c.a = NULL;
	return 0;
}

/*! @ingroup PDBlib
 * @brief Add new atom object.
 * @param PDBh - initialized PDBlib handler
 * @return Functions returns pointer to PDBlib_atom_t structure.
 *
 * Function creates new atom object. If needed top-level objects
 * are created as well. Newly created atom object is set as
 * referencing point.
 */
PDBlib_atom_t *PDBlib_add_atom(PDBhandle PDBh)
{
	PDBlib_atom_t new_a;

	// if current residue object is NULL add new one
	if(PDBh->c.r == NULL) PDBh->c.r = PDBlib_add_residue(PDBh);

	bzero(&new_a, sizeof(new_a));
	new_a.id = random() + 1;
	new_a.top_residue = PDBh->c.r;

	g_array_append_vals(PDBh->c.r->atoms, &new_a, 1);
	return PDBh->c.a = &g_array_index(PDBh->c.r->atoms, PDBlib_atom_t,
			PDBh->c.r->atoms->len - 1);
}

/*! @ingroup PDBlib
 * @brief Get atom object.
 * @param PDBh - initialized PDBlib handler
 * @param id - unique atom ID
 * @return On success pointer to PDBlib_atom_t structure is returned.
 *
 * For more information see PDBlib_get_structure().
 */
PDBlib_atom_t *PDBlib_get_atom(PDBhandle PDBh, int id)
{
	PDBlib_atom_t *cur_a;
	int xa;

	if(PDBh->c.r == NULL) return NULL;
	for(xa = 0; xa < PDBh->c.r->atoms->len; xa++){
		cur_a = &g_array_index(PDBh->c.r->atoms, PDBlib_atom_t, xa);
		if(cur_a->id == id) return PDBh->c.a = cur_a;}
	return NULL;
}

/*! @ingroup PDBlib
 * @brief Delete atom object.
 * @param PDBh - initialized PDBlib handler
 * @param id - unique atom ID
 * @return On success 0 is returned, otherwise -1.
 *
 * For more information see PDBlib_del_structure().
 */
int PDBlib_del_atom(PDBhandle PDBh, int id)
{
	PDBlib_atom_t *cur_a = NULL;
	int xa;

	if(PDBh->c.r == NULL) return -1;
	for(xa = 0; xa < PDBh->c.r->atoms->len; xa++){
		cur_a = &g_array_index(PDBh->c.r->atoms, PDBlib_atom_t, xa);
		if(cur_a->id == id) break;}

	// residue with requested ID not found
	if(xa == PDBh->c.r->atoms->len) return -1;

	// if needed update hetero_counter in top_residue
	if(cur_a->hetero == 1) cur_a->top_residue->hetero_counter--;

	g_array_remove_index(PDBh->c.r->atoms, xa);

	// update ID for remaining atoms
	for(; xa < PDBh->c.r->atoms->len; xa++){
		cur_a = &g_array_index(PDBh->c.r->atoms, PDBlib_atom_t, xa);
		cur_a->id = xa;}

	PDBh->c.a = NULL;
	return 0;
}

/*! @ingroup PDBlib
 * @brief Initialize main data memory block.
 * @return On success returns initialized PDBlib handler, otherwise NULL.
 *
 * This function allocate memory for PDBlib handler object and prepares
 * it to use. Returned pointer has to be freed by call to PDBlib_free()
 * function.
 */
PDBhandle PDBlib_initialize()
{
	PDBhandle PDBh;

	// initialize main data memory block
	if((PDBh = (PDBhandle)malloc(sizeof(struct PDBlib_maindata))) == NULL)
		return NULL;
	bzero(PDBh, sizeof(struct PDBlib_maindata));

	// initialize structures object
	PDBh->structures = g_array_new(FALSE, TRUE, sizeof(PDBlib_struct_t));
	return PDBh;
}

/*! @ingroup PDBlib
 * @brief Free memory associated with PDB handler.
 * @param PDBh - initialized PDBlib handler
 *
 * This function clean memory which has been allocated by
 * PDBlib_initialize() and by various PDBlib_* functions.
 */
void PDBlib_free(PDBhandle PDBh)
{
	PDBlib_struct_t *cur_s;
	int xs;

	// free whole PDB structure tree
	for(xs = 0; xs < PDBh->structures->len; xs++){
		cur_s = &g_array_index(PDBh->structures, PDBlib_struct_t, xs);
		PDBlib_del_structure(PDBh, cur_s->id);}
	g_array_free(PDBh->structures, TRUE);

	free(PDBh);
}
