#include <string.h>
#include <ctype.h>
#include "pdblib.h"

static int _checkChainsNaming(PDBhandle PDBh, PDBlib_iter_data_t *data);
static int _guessAtomElement(PDBhandle PDBh, PDBlib_iter_data_t *data);
static int _serialNbRenumbering(PDBhandle PDBh, PDBlib_iter_data_t *data);
static gint _compareAtoms(gconstpointer a, gconstpointer b);
static int _sortAtomsInResidues(PDBhandle PDBh, PDBlib_iter_data_t *data);

char *strstrip(char *s)
{
	int x;
	for(x = 0; x < strlen(s); x++) if(!isspace(s[x])) break;
	memmove(s, &s[x], strlen(&s[x]) + 1);
	for(x = strlen(s); x > 0; x--) if(!isspace(s[x-1])) break;
	s[x] = 0;
	return s;
}

/// @cond HIDDEN_SYMBOLS
struct _optimization_data {
	int serial_nb_counter;
	int prev_chain_id, prev_model_id;
};
/// @endcond

/*! @ingroup PDBlib
  @brief Perform optimization on PDB structure.
  @param PDBh - initialized PDBlib handler
  @param flags - optimization flag(s)

  This optimization may includes:
  @n + sorting atoms in residue (order: N, CA, C, O, alphabetically)
  @n + correction in atom->serial_nb numbering
  @n + adding element field if omitted (by guessing it from atom name)

  Available flags (can be combined with "|"):
  @n + PDBLIB_OPTI_SORTATOMS
  @n + PDBLIB_OPTI_RENUMBERING
  @n + PDBLIB_OPTI_ADDELEMENT
  @n + PDBLIB_OPTI_ADDCHAINAME
  @n + PDBLIB_OPTI_ALL 
 */
void PDBlib_perform_optimization(PDBhandle PDBh, int flags)
{
	struct _optimization_data optd;
	int counter;

	bzero(&optd, sizeof(optd));
	optd.serial_nb_counter = 1; //start numbering from 1
	counter = 0;

	if((flags & PDBLIB_OPTI_SORTATOMS) != 0)
		PDBlib_iterator(PDBh, PDBLIB_ITER_RESIDUE, _sortAtomsInResidues, NULL);
	if((flags & PDBLIB_OPTI_RENUMBERING) != 0)
		PDBlib_iterator(PDBh, PDBLIB_ITER_ATOM, _serialNbRenumbering, &optd);
	if((flags & PDBLIB_OPTI_ADDELEMENT) != 0)
		PDBlib_iterator(PDBh, PDBLIB_ITER_ATOM, _guessAtomElement, NULL);
	if((flags & PDBLIB_OPTI_ADDCHAINAME) != 0)
		PDBlib_iterator(PDBh, PDBLIB_ITER_CHAIN, _checkChainsNaming, &counter);
}

// If chain contains only heteroatoms ist name is set to blank. Elsewhere
// if chains name is blank it is set to A, B, C, ...
static int _checkChainsNaming(PDBhandle PDBh, PDBlib_iter_data_t *data) {
	PDBlib_residue_t *cur_r;
	int xr, *counter = (int*)data->userdata;

	// check if all atoms in chain are heteroatoms
	// if YES set chain name to blank
	for(xr = 0; xr < data->c->residues->len; xr++){
		cur_r = &g_array_index(data->c->residues, PDBlib_residue_t, xr);
		if(cur_r->atoms->len != cur_r->hetero_counter) break;}

	if(xr == data->c->residues->len) data->c->name[0] = 0;
	else {
		// if there is already chain name leave it
		if(data->c->name[0] != 0 && !isspace(data->c->name[0])) return 0;

		data->c->name[0] = *counter + 'A';
	}

	(*counter)++;
	return 0;
}

// Assign element (if not assigned) as first two letters form atom name field.
static int _guessAtomElement(PDBhandle PDBh, PDBlib_iter_data_t *data) {
	char elem_strip[3];

	strstrip(strcpy(elem_strip, data->a->element));
	if(strlen(elem_strip) != 0) return 0;

	bzero(data->a->element, sizeof(data->a->element));
	memcpy(data->a->element, data->a->name, 2);

	// fix common situations
	if(memcmp(data->a->element, "HH" , 2) == 0 ||
			memcmp(data->a->element, "HD" , 2) == 0)
		data->a->element[1] = 0;
	return 0;
}

// Re-numerate a->serial_nb (starting from 1; see PDBlib_perform_optimization)
static int _serialNbRenumbering(PDBhandle PDBh, PDBlib_iter_data_t *data) {
	struct _optimization_data *optd = (struct _optimization_data*)data->userdata;

	if(optd->prev_chain_id != data->c->id)
		optd->serial_nb_counter++; //reserv number for TER record

	if(optd->prev_model_id != data->m->id)
		optd->serial_nb_counter = 1; //new model -> start counting from 1

	data->a->serial_nb = optd->serial_nb_counter++;
	optd->prev_model_id = data->m->id;
	optd->prev_chain_id = data->c->id;
	return 0;
}

// Atom comparison function
// Sorting order: N, CA, C, O, alphabetically
static gint _compareAtoms(gconstpointer a, gconstpointer b) {
	PDBlib_atom_t *a1 = (PDBlib_atom_t*)a;
	PDBlib_atom_t *a2 = (PDBlib_atom_t*)b;
	char *pro_order[] = {"N", "CA", "C", "O"};
	char sname1[5], sname2[5];
	int x, pro1, pro2;

	// strip spaces from atoms names
	strstrip(strcpy(sname1, a1->name));
	strstrip(strcpy(sname2, a2->name));

	if(strcmp(sname1, sname2) == 0)
		return memcmp(&a1->alt_location, &a2->alt_location, sizeof(a1->alt_location));

	for(x = pro1 = pro2 = 0; x < sizeof(pro_order)/sizeof(char*); x++){
		if(strcmp(sname1, pro_order[x]) == 0) pro1 = x + 1;
		if(strcmp(sname2, pro_order[x]) == 0) pro2 = x + 1;}

	if(pro1 != 0 && pro2 != 0) return memcmp(&pro1, &pro2, sizeof(pro1));
	if(pro1 != 0) return -1;
	if(pro2 != 0) return 1;
	return strcmp(sname1, sname2);
}
static int _sortAtomsInResidues(PDBhandle PDBh, PDBlib_iter_data_t *data) {
	PDBlib_atom_t *cur_a;
	int xa;

	g_array_sort(data->r->atoms, _compareAtoms);

	// update atoms IDs
	for(xa = 0; xa < data->r->atoms->len; xa++){
		cur_a = &g_array_index(data->r->atoms, PDBlib_atom_t, xa);
		cur_a->id = xa;}

	return 0;
}
