/** @defgroup PDBlib PDBlib
 
   Library providing basic functionality for accessing PDB files.
 */

#include <time.h>
#include <glib.h>

#include <stdio.h>
#define debug(s) printf(s); fflush(stdout)
#define debugg(s, d) printf(s,d); fflush(stdout)

// PDB objects hierarchy: structure->model->chain->residue->atom

#define INTERNAL_NAME __attribute__ ((visibility("hidden")))

/*! @ingroup PDBlib 
 */
typedef struct PDBlib_structdata_tag {
	int id;

	// HEADER related data
	char classification[41], PDBID[5];
	struct tm depos_data;

	GArray *models;
	GArray *ssbonds;
}PDBlib_struct_t;

/*! @ingroup PDBlib
 */
typedef struct PDBlib_modeldata_tag {
	int id;
	PDBlib_struct_t *top_structure;

	int serial_nb;
	GArray *chains;
}PDBlib_model_t;

/*! @ingroup PDBlib
 */
typedef struct PDBlib_chaindata_tag {
	int id;
	PDBlib_model_t *top_model;

	char name[2];
	GArray *residues;
}PDBlib_chain_t;

/*! @ingroup PDBlib
 */
typedef struct PDBlib_residuedata_tag {
	int id;
	PDBlib_chain_t *top_chain;

	int sequence_nb, hetero_counter;
	char name[4], segment[5], insertion;
	GArray *atoms;
}PDBlib_residue_t;

/*! @ingroup PDBlib
 */
typedef struct PDBlib_atomdata_tag {
	int id;
	PDBlib_residue_t *top_residue;

	int serial_nb, hetero;
	double coordXYZ[3], occupancy, bfactor;
	char name[5], element[3], charge[3], alt_location;
	int aniso_temp_factor[6]; //{u11, u22, u33, u12, u13, u23}
}PDBlib_atom_t;

/*! @ingroup PDBlib
 */
typedef struct PDBlib_ssbonddata_tag {
	int id;
	PDBlib_struct_t *top_structure;

	int serial_nb, resseq_nb1, resseq_nb2, sym_oper1, sym_oper2;
	char chainame1[2], chainame2[2], res_insertion1, res_insertion2;
	double sslen;
}PDBlib_ssbond_t;

/// @cond HIDDEN_SYMBOLS
struct PDBlib_maindata {
	GArray *structures;

	struct current_obj { //used by object module
		PDBlib_struct_t *s;
		PDBlib_model_t *m;
		PDBlib_chain_t *c;
		PDBlib_residue_t *r;
		PDBlib_atom_t *a;
		PDBlib_ssbond_t *sb;
	}c;

	struct previous_obj_id { //used by parser and save module
		int s, m, c, r, a, ss;
	}pi;
};
/// @endcond
typedef struct PDBlib_maindata* PDBhandle;


// stuff from pdbparser.c
int PDBlib_load_file(PDBhandle PDBh, const char *fname, int pqr_ext);
int PDBlib_save_file(PDBhandle PDBh, const char *fname, int pqr_ext);

// stuff from functions.c
void PDBlib_remove_water(PDBhandle PDBh);
void PDBlib_remove_heteroatoms(PDBhandle PDBh);

int PDBlib_set_residue_hashed_id(PDBlib_residue_t *src_r);
int PDBlib_set_atom_hashed_id(PDBlib_atom_t *src_a);

void PDBlib_copy_atom_data(PDBlib_atom_t *dst, const PDBlib_atom_t *src);
PDBhandle PDBlib_duplicate(const PDBhandle PDBh);

// stuff from optimize.c
#define PDBLIB_OPTI_SORTATOMS   0x0001
#define PDBLIB_OPTI_RENUMBERING 0x0002
#define PDBLIB_OPTI_ADDELEMENT  0x0004
#define PDBLIB_OPTI_ADDCHAINAME 0x0008
#define PDBLIB_OPTI_ALL         0x000F
void PDBlib_perform_optimization(PDBhandle PDBh, int flags);

// stuff from iterator.c
#define PDBLIB_ITER_ATOM      1
#define PDBLIB_ITER_RESIDUE   2
#define PDBLIB_ITER_CHAIN     3
#define PDBLIB_ITER_MODEL     4
#define PDBLIB_ITER_STRUCTURE 5
#define PDBLIB_ITER_SSBOND    6
/*! @ingroup PDBlib
 */
typedef struct PDBlib_iter_data_tag {
	PDBlib_struct_t *s;
	PDBlib_model_t *m;
	PDBlib_chain_t *c;
	PDBlib_residue_t *r;
	PDBlib_atom_t *a;
	PDBlib_ssbond_t *sb;
	void *userdata;
}PDBlib_iter_data_t;

/*! @ingroup PDBlib
 */
typedef int (*PDBlib_iter_callback_t)(PDBhandle PDBh, PDBlib_iter_data_t *data);

void PDBlib_iterator(PDBhandle PDBh, int type, PDBlib_iter_callback_t callback, void *userdata);

// stuff from objects.c
PDBhandle PDBlib_initialize();
void PDBlib_free(PDBhandle PDBh);

PDBlib_struct_t *PDBlib_add_structure(PDBhandle PDBh);
PDBlib_model_t *PDBlib_add_model(PDBhandle PDBh);
PDBlib_chain_t *PDBlib_add_chain(PDBhandle PDBh);
PDBlib_residue_t *PDBlib_add_residue(PDBhandle PDBh);
PDBlib_atom_t *PDBlib_add_atom(PDBhandle PDBh);
PDBlib_ssbond_t *PDBlib_add_ssbond(PDBhandle PDBh);

PDBlib_struct_t *PDBlib_get_structure(PDBhandle PDBh, int id);
PDBlib_model_t *PDBlib_get_model(PDBhandle PDBh, int id);
PDBlib_chain_t *PDBlib_get_chain(PDBhandle PDBh, int id);
PDBlib_residue_t *PDBlib_get_residue(PDBhandle PDBh, int id);
PDBlib_atom_t *PDBlib_get_atom(PDBhandle PDBh, int id);
PDBlib_ssbond_t *PDBlib_get_ssbond(PDBhandle PDBh, int id);

#define PDBlib_chain_get_res(c, r) &g_array_index((c)->residues, PDBlib_residue_t, (r))
#define PDBlib_res_get_atom(r, i) &g_array_index((r)->atoms, PDBlib_atom_t, (i))

int PDBlib_del_structure(PDBhandle PDBh, int id);
int PDBlib_del_model(PDBhandle PDBh, int id);
int PDBlib_del_chain(PDBhandle PDBh, int id);
int PDBlib_del_residue(PDBhandle PDBh, int id);
int PDBlib_del_atom(PDBhandle PDBh, int id);
int PDBlib_del_ssbond(PDBhandle PDBh, int id);

